SystemOrganization addCategory: #TwosComplement!

Magnitude subclass: #TwosComplementRegister
	instanceVariableNames: 'bits overflow'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TwosComplement'!

!TwosComplementRegister commentStamp: 'dtl 9/30/2011 14:43' prior: 0!
TwosComplementRegister implements a twos complement register of arbitrary width.
It consists of an array of bits (boolean values) and methods to perform bit arithmetic.
A register accumulates the results of successive operations, and maintains an overflow
bit to indicate status of the preceding operation.

The current state of a register may represent either an unsigned integer or a signed
integer. In either case, the register operations (addition, subtraction) are identical. When
interpreted as a signed integer, the high order bit of the register is treated as the sign
bit, and the integer value is determined by the remaining bits.

A register may be of any width, although powers of two (8, 16, 32 ...) will match the
expectations of data types in conventional programming languages.

All operations are implemented in terms of boolean logic without reliance on numbers.

Instance Variables
	bits:		array of boolean bits, of size equal to register width
	overflow:	true if operation that set value resulted in arithmetic overflow!

!TwosComplementRegister class methodsFor: 'instance creation' stamp: 'dtl 9/29/2011 08:08'!
value: anInteger
	"default to a 32-bit register size"
	^self new width: 32 value: anInteger! !

!TwosComplementRegister class methodsFor: 'instance creation' stamp: 'dtl 9/29/2011 20:49'!
width: bits value: anInteger
	"Answer a new register of the given width initialized to an integer value.
	The value may be a signed or unsigned integer, and will be stored in twos
	complement form with overflow set if the value exceeds the range of the
	register. A twos complement register of width 8 can store any integer value
	in the range -128 to 255 without incurring overflow."

	^self new width: bits value: anInteger! !

!TwosComplementRegister methodsFor: 'comparing' stamp: 'dtl 9/29/2011 08:03'!
< aMagnitude 
	"Answer whether the receiver is less than the argument."

	^self subclassResponsibility! !

!TwosComplementRegister methodsFor: 'comparing' stamp: 'dtl 9/30/2011 08:08'!
= aRegister 
	"Compare the receiver with the argument and answer with true if the 
	receiver is equal to the argument. Otherwise answer false."

	^bits = aRegister bits and: [overflow = aRegister overflow]! !

!TwosComplementRegister methodsFor: 'register arithmetic' stamp: 'dtl 9/30/2011 08:19'!
add: aRegister
	"Add the contents of aRegister to this register"

	| rs last |
	rs := ReadStream on: aRegister bits.
	overflow :=  (1 to: bits size)
		inject: false into: [:carry :i | | thisBit thatBit |
			thisBit := bits at: i.
			rs atEnd
				ifFalse: [last := thatBit := rs next]
				ifTrue: [thatBit := last"extend the sign bit"] .
			"value of this bit is the XOR of the carry and two operand bits"
			bits at: i put: ((thisBit xor: thatBit) xor: carry).
			"if two or more of the three bits are set, perform a carry"
			({ thisBit . thatBit . carry } copyWithout: false) size > 1].
	(overflow or: [rs atEnd])
		ifFalse: ["If there are more bits to evaluate in the other register, check for overflow.
			The remaining bits must be the same as last, either all ones or all zeros"
			[rs atEnd] whileFalse: [overflow := overflow or: [(rs next = last) not]]]
! !

!TwosComplementRegister methodsFor: 'integer arithmetic' stamp: 'dtl 9/30/2011 08:20'!
addInteger: anInteger

	| reg |
	reg := self species width: bits size value: anInteger.
	self add: reg
! !

!TwosComplementRegister methodsFor: 'converting' stamp: 'dtl 9/29/2011 20:41'!
asInteger
	^self asSignedInteger
! !

!TwosComplementRegister methodsFor: 'converting' stamp: 'dtl 9/29/2011 20:42'!
asSignedInteger

	self negative
		ifTrue: [^ ((bits reversed collect: [:e | e not])
				inject: 0
				into: [:val :bit | bit
						ifTrue: [val << 1 + 1]
						ifFalse: [val << 1]]) negated - 1]
		ifFalse: [^ self asUnsignedInteger]! !

!TwosComplementRegister methodsFor: 'converting' stamp: 'dtl 9/29/2011 20:41'!
asUnsignedInteger
	^ bits reversed
				inject: 0
				into: [:val :bit | bit
						ifTrue: [val << 1 + 1]
						ifFalse: [val << 1]]! !

!TwosComplementRegister methodsFor: 'accessing' stamp: 'dtl 9/29/2011 18:32'!
bitAt: position
	"Answer the value of the bit at position, where position 1 is the least significant bit"

	^ bits at: position! !

!TwosComplementRegister methodsFor: 'printing' stamp: 'dtl 9/29/2011 18:38'!
bitPatternString
	^ String streamContents: [:strm |
			bits reverseDo: [:bit | strm nextPut: (bit ifTrue: [$1] ifFalse: [$0])]]! !

!TwosComplementRegister methodsFor: 'accessing' stamp: 'dtl 9/29/2011 19:04'!
bits
	^bits! !

!TwosComplementRegister methodsFor: 'comparing' stamp: 'dtl 9/30/2011 08:10'!
hash
	"Hash must be redefined whenever = is redefined."

	^overflow == true
		ifTrue: [bits hash + 1]
		ifFalse: [bits hash]! !

!TwosComplementRegister methodsFor: 'register arithmetic' stamp: 'dtl 9/30/2011 08:19'!
increment

	overflow :=  (1 to: bits size)
		inject: true into: [:carry :i | | bit |
			bit := bits at: i.
			bits at: i put: (bit xor: carry).
			bit and: carry].
		! !

!TwosComplementRegister methodsFor: 'testing' stamp: 'dtl 9/30/2011 16:10'!
maxNegative
	"true if register contents represent a maximum signed negative value for
	this register size, for which no corresponding positive value exists."
	^self negative and: [(bits allButLast includes: true) not]
! !

!TwosComplementRegister methodsFor: 'register arithmetic' stamp: 'dtl 9/30/2011 16:14'!
negate
	"Assume that the register contents are to be interpreted as a signed
	integer value. Perform twos complement negations, and set the overflow
	bit if and only if value is maximum negative (all zeros except sign bit).
	Negation of a maximum negative value is a special case because there
	is no corresponding positive value for that magnitude. Twos complement
	of the maximum negative results in the identical register value, and
	overflow is set to indicate that the resulting value is undefined."

	| willOverflow |
	willOverflow := self maxNegative.
	self twosComplement.
	overflow := willOverflow! !

!TwosComplementRegister methodsFor: 'testing' stamp: 'dtl 9/30/2011 08:20'!
negative
	"high order bit is the sign bit"
	^bits at: bits size! !

!TwosComplementRegister methodsFor: 'initialize-release' stamp: 'dtl 9/29/2011 08:13'!
normalize
	"check for overflow and trim value to fit register size"! !

!TwosComplementRegister methodsFor: 'register arithmetic' stamp: 'dtl 9/30/2011 16:28'!
onesComplement
	"Perform a ones complement operation (invert bits) on the current register
	contents. In ones complement arithmetic, both positive and negative zero
	exist. In twos complement, negation is a ones complement followed by an
	increment to eliminate the negative zero."

	(1 to: bits size) do: [:i |
		bits at: i put: (bits at: i) not]
! !

!TwosComplementRegister methodsFor: 'testing' stamp: 'dtl 9/29/2011 19:00'!
overflow
	"true if the last operation on this register value resulted in arithmetic overflow"
	^overflow! !

!TwosComplementRegister methodsFor: 'printing' stamp: 'dtl 9/30/2011 08:25'!
printOn: aStream
	"Append to the argument, aStream, a sequence of characters that  
	identifies the receiver."
	super printOn: aStream.
	aStream nextPutAll: (overflow == true ifTrue: [' {'] ifFalse: [' (']).
	bits reverseDo: [:bit | aStream nextPut: (bit ifTrue: [$1] ifFalse: [$0])].
	aStream nextPut: (overflow == true ifTrue: [$}] ifFalse: [$)]).
! !

!TwosComplementRegister methodsFor: 'initialize-release' stamp: 'dtl 9/29/2011 19:55'!
reset
	overflow := false.
	bits atAllPut: false
! !

!TwosComplementRegister methodsFor: 'initialize-release' stamp: 'dtl 9/30/2011 08:19'!
setValue: anInteger

	| val |
	val := anInteger.
	(1 to: bits size) do: [:i |
		bits at: i put: (val bitAnd: 1) = 1.
		val := val >> 1].
	overflow := self negative
		ifTrue: [val ~= -1 and: (val ~= 0)]
		ifFalse: [val ~= 0]
! !

!TwosComplementRegister methodsFor: 'register arithmetic' stamp: 'dtl 9/30/2011 17:30'!
subtract: aRegister
	"Subtract the contents of aRegister from this register"

	| isZero |
	isZero := self zero.
	self add: aRegister copy negate.
	(isZero and: [self maxNegative])
		ifTrue: [overflow := true]

! !

!TwosComplementRegister methodsFor: 'integer arithmetic' stamp: 'dtl 9/30/2011 17:37'!
subtractInteger: anInteger

	| reg |
	reg := self species width: bits size value: anInteger.
	self subtract: reg
! !

!TwosComplementRegister methodsFor: 'register arithmetic' stamp: 'dtl 9/30/2011 16:26'!
twosComplement
	"Perform a twos complement operation on the current register contents"

	overflow := false.
	self onesComplement.
	self increment
! !

!TwosComplementRegister methodsFor: 'initialize-release' stamp: 'dtl 9/30/2011 08:20'!
width: size value: anInteger
	bits := Array new: size withAll: false.
	self setValue: anInteger
! !

!TwosComplementRegister methodsFor: 'testing' stamp: 'dtl 9/30/2011 16:13'!
zero
	"true if register contents represent zero, valid for signed or unsigned integers.
	No negative zero value is defined."
	^(bits includes: true) not
! !

TestCase subclass: #TwosComplementRegisterTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TwosComplement'!

!TwosComplementRegisterTest commentStamp: 'dtl 9/30/2011 17:35' prior: 0!
A TwosComplementRegisterTest documents the expected behavior of a TwosComplementRegister
!

!TwosComplementRegisterTest methodsFor: 'testing' stamp: 'dtl 9/30/2011 17:28'!
testAccumulator
	"A register is not a number, it is an accumulator"
	| reg val |
	reg := TwosComplementRegister width: 13 value: 0.
	val := (0 to: 1000 by: 3) inject: 0 into: [ :currentValue :nextValue |
		self assert: reg asInteger = currentValue.
		reg addInteger: 3.
		nextValue + 3].
	(val to: -2000 by: -5) inject: val into: [ :currentValue :nextValue |
		self assert: reg asInteger = currentValue.
		reg addInteger: -5.
		nextValue - 5].! !

!TwosComplementRegisterTest methodsFor: 'testing' stamp: 'dtl 9/30/2011 14:45'!
testAdd

	| reg1 reg2 |
	(-128 to: 127) do: [ :i |
		reg1 := TwosComplementRegister width: 8 value: 0.
		reg2 := TwosComplementRegister width: 8 value: i.
		reg1 add: reg2.
		self assert: reg1 asInteger = i.
		self deny: reg1 overflow].

	(-128 to: 127) do: [ :i |
		reg1 := TwosComplementRegister width: 8 value: i.
		reg2 := TwosComplementRegister width: 8 value: 0.
		reg1 add: reg2.
		self assert: reg1 asInteger = i.
		self deny: reg1 overflow].

	"test 8 bit register, skip some values for speed"
	(-128 to: 127 by: 5) do: [ :i | | start end |
		start := -128 - i.
		end := 127 - i.
		(start to: end by: 5) do: [ :k |
			reg1 := TwosComplementRegister width: 8 value: i.
			reg2 := TwosComplementRegister width: 8 value: k.
			reg1 add: reg2.
			self assert: i + k = reg1 asInteger.
			"self deny: reg1 overflow"  "overflow expected for many combinations"]].

	"verify full range of 5 bit register with unsigned values"
	(0 to: 31) do: [ :i |
		(0 to: 31) do: [ :k | | expect |
			reg1 := TwosComplementRegister width: 5 value: i.
			reg2 := TwosComplementRegister width: 5 value: k.
			reg1 add: reg2.
			expect :=  (i + k) \\ 32.
			self assert: expect = reg1 asUnsignedInteger.
			reg1 overflow "overflow bit indicates rollover occurred"
				ifTrue: [self assert: expect = (i + k - 32)]
				ifFalse: [self assert: expect = (i + k)]]].

	"Test 5 bit register, cover full range, signed arithmetic. The arithmetic
	results will be as expected for signed arithmetic, and the overflow bit
	will be set if overflow occurred (though this does not affect validity of
	the signed arithmetic result)."
	(-16 to: 15) do: [ :i | | start end |
		start := (-16 - i) max: -16.
		end := (15 - i) min: 15.
		(start to: end) do: [ :k |
			reg1 := TwosComplementRegister width: 5 value: i.
			reg2 := TwosComplementRegister width: 5 value: k.
			reg1 add: reg2.
			self assert: i + k = reg1 asInteger.
			"self deny: reg1 overflow"  "overflow expected for many combinations"]].
! !

!TwosComplementRegisterTest methodsFor: 'testing' stamp: 'dtl 9/29/2011 21:13'!
testAsInteger

	| reg |
	"normal range of signed integers, no overflow"
	(-128 to: 127) do: [ :i |
		reg := TwosComplementRegister width: 8 value: i.
		self assert: reg asInteger = i]
! !

!TwosComplementRegisterTest methodsFor: 'testing' stamp: 'dtl 9/29/2011 21:14'!
testAsUnsignedInteger

	| reg |
	(0 to: 255) do: [ :i |
		reg := TwosComplementRegister width: 8 value: i.
		self assert: reg asUnsignedInteger = i]
! !

!TwosComplementRegisterTest methodsFor: 'testing' stamp: 'dtl 9/29/2011 22:01'!
testIncrement

	| reg |
	(0 to: 254) do: [ :i |
		reg := TwosComplementRegister width: 8 value: i.
		reg increment.
		self deny: reg overflow.
		self assert: (i + 1 = reg asUnsignedInteger)].
	reg := TwosComplementRegister width: 8 value: 255.
	reg increment.
	self assert: reg overflow.
	self assert: reg asInteger = 0.

	(-128 to: -2) do: [ :i |
		reg := TwosComplementRegister width: 8 value: i.
		reg increment.
		self deny: reg overflow.
		self assert: (i + 1 = reg asInteger)].
	reg := TwosComplementRegister width: 8 value: -1.
	reg increment.
	self assert: reg overflow.
	self assert: reg asInteger = 0.
! !

!TwosComplementRegisterTest methodsFor: 'testing' stamp: 'dtl 9/29/2011 22:06'!
testNegate

	| reg |
	(0 to: 127) do: [ :i |
		reg := TwosComplementRegister width: 8 value: i.
		reg negate.
		self deny: reg overflow.
		self assert: (reg asInteger = i negated)].
	(-127 to: -1) do: [ :i |
		reg := TwosComplementRegister width: 8 value: i.
		reg negate.
		self deny: reg overflow.
		self assert: (reg asInteger = i negated)].

	"In C language, negation results are undefined for maximum negative, but this is the behavior for a real register"
	reg := TwosComplementRegister width: 8 value: -128.
	reg negate.
	self assert: reg overflow.
	self assert: (reg asInteger = -128)
! !

!TwosComplementRegisterTest methodsFor: 'testing' stamp: 'dtl 9/29/2011 21:18'!
testSetOverflowOnInitialize

	| reg |
	"normal range of signed integers, no overflow"
	(-128 to: 127) do: [ :i |
		reg := TwosComplementRegister width: 8 value: i.
		self deny: reg overflow].
	"overflow values are still valid when treated as unsigned, so do not set the overflow bit"
	(128 to: 255) do: [ :i |
		reg := TwosComplementRegister width: 8 value: i.
		self deny: reg overflow].
	"out of range, set overflow bit"
	(-10000 to: -129) do: [ :i |
		reg := TwosComplementRegister width: 8 value: i.
		self assert: reg overflow].
	(257 to: 10000) do: [ :i |
		reg := TwosComplementRegister width: 8 value: i.
		self assert: reg overflow]
! !

!TwosComplementRegisterTest methodsFor: 'testing' stamp: 'dtl 9/30/2011 17:20'!
testSubtract

	| reg1 reg2 |
	"zero minus max negative is a special case, exceeds positive range"
	reg1 := TwosComplementRegister width: 8 value: 0.
	reg2 := TwosComplementRegister width: 8 value: -128.
	reg1 subtract: reg2.
	self assert: reg1 asInteger =  -128.
	self assert: reg1 asUnsignedInteger = 128.
	self assert: reg1 overflow. "result as signed integer is in undefined positive range"
	(-127 to: 127) do: [ :i |
		reg1 := TwosComplementRegister width: 8 value: 0.
		reg2 := TwosComplementRegister width: 8 value: i.
		reg1 subtract: reg2.
		self assert: reg1 asInteger = i negated.
		self deny: reg1 overflow].

	(-128 to: 127) do: [ :i |
		reg1 := TwosComplementRegister width: 8 value: i.
		reg2 := TwosComplementRegister width: 8 value: 0.
		reg1 subtract: reg2.
		self assert: reg1 asInteger = i.
		self deny: reg1 overflow].

	"test 5 bit register, cover full range"
	(-16 to: 15) do: [ :i | | start end |
		start := 16 - i.
		end := i - 15.
		(start to: end) do: [ :k |
			reg1 := TwosComplementRegister width: 5 value: i.
			reg2 := TwosComplementRegister width: 5 value: k.
			reg1 subtract: reg2.
			self assert: i - k = reg1 asInteger.
			self deny: reg1 overflow]].

	"test 8 bit register, skip some values for speed"
	(-128 to: 127 by: 5) do: [ :i | | start end |
		start := 128 - i. "n.b. skip value -128 which will overflow"
		end := i - 127.
		(start to: end by: 5) do: [ :k |
			reg1 := TwosComplementRegister width: 8 value: i.
			reg2 := TwosComplementRegister width: 8 value: k.
			reg1 subtract: reg2.
			self assert: (i + k) negated = reg1 asInteger.
			self deny: reg1 overflow]].
! !
